from dataclasses import dataclass

import tables as tb
import time
import numpy as np
from typing import Dict, Tuple, Union, List
import os
from enum import Enum, auto


@dataclass
class StateData:
    observation: Tuple[np.ndarray]
    additional_data: Dict[str, np.ndarray]
    terminal: bool


LOGS_GROUP = 'logs'
LOG_ID_GROUP = 'log_id'
TRAJECTORIES_ARRAY = 'trajectories_start_idx'
LOG_STATES_GROUP = 'states'
STATE_ELEM = 'state_elem_'
LOG_ACTIONS_ARRAY = 'actions'
LOG_REWARDS_ARRAY = 'rewards'


DB_STRUCTURE = {
    LOGS_GROUP: {
        LOG_ID_GROUP: {

            LOG_STATES_GROUP: {
                STATE_ELEM: np.ndarray
            }
        }
    }
}


class PosInTrajectory(Enum):
    FIRST = auto()
    LAST = auto()
    MID = auto()
    ONLY = auto()


class DBWriter:
    """
    Intended for easy to use interface with the DB to perform the necessary DB write transactions.
    An object of this class records all interactions as a single run.
    To record different runs, use different objects.

    * avoid messy function signatures by avoiding to require configurations in function params

    """
    def __init__(
            self,
            state_shape: Union[Tuple[int], Tuple[Tuple[int]]],
            action_shape: Tuple[int],
            is_action_discrete: bool = True,
            db_file: str = 'monitor_db.h5',
            total_steps_estimate: int = None
    ):
        self.state_shape = state_shape
        if isinstance(self.state_shape[0], int):
            self.state_shape = (self.state_shape,)
        self.action_shape = action_shape
        self.is_action_discrete = is_action_discrete
        self.total_steps_estimate = total_steps_estimate

        self.db_file = db_file
        open_mode = 'a' if os.path.exists(self.db_file) else 'w'
        self.db = tb.open_file(self.db_file, mode=open_mode, title="RL Monitor DB")

        self.run_id = self._generate_run_id()
        self._initialize_monitor_run()

        self._start_new_trajectory_flag = False

    def __del__(self):
        self.db.close()

    def store_transition(
            self,
            state: Union[np.ndarray, Tuple[np.ndarray]],
            action,
            reward: float,
            next_state: Union[np.ndarray, Tuple[np.ndarray]],
            info = None,
            pos_in_trajectory: PosInTrajectory = PosInTrajectory.MID
    ):
        # If this state begins a new trajectory, store that information as well:
        if pos_in_trajectory == PosInTrajectory.FIRST or pos_in_trajectory == PosInTrajectory.ONLY:
            if self._start_new_trajectory_flag:
                self._begin_new_trajectory()
            self._store_state(state)
        self._start_new_trajectory_flag = False

        # self._store_state(state)
        self._store_action(action)
        self._store_reward(reward)
        self._store_state(next_state)

        if pos_in_trajectory == PosInTrajectory.LAST or pos_in_trajectory == PosInTrajectory.ONLY:
            self._start_new_trajectory_flag = True

    def _store_state(self, env_state: Union[np.ndarray, Tuple[np.ndarray]]):
        # wrap single obs states in a tuple for uniform handling:
        if not isinstance(env_state, tuple):
            assert isinstance(env_state, np.ndarray)
            env_state = (env_state,)

        states_group_node = self.db.get_node(f'/{LOGS_GROUP}/{self.run_id}/{LOG_STATES_GROUP}')
        assert states_group_node._v_nchildren == len(env_state)

        for i, state_elem in enumerate(env_state):
            state_elem_arr = self.db.get_node(f'/{LOGS_GROUP}/{self.run_id}/{LOG_STATES_GROUP}/{STATE_ELEM}{i}')
            state_elem_arr.append(state_elem)

    def _store_action(self, agent_action):
        if isinstance(agent_action, (int, float)):
            agent_action = np.array([agent_action])
        actions_arr = self.db.get_node(f'/{LOGS_GROUP}/{self.run_id}/{LOG_ACTIONS_ARRAY}')
        actions_arr.append(agent_action)

    def _store_reward(self, reward: float):
        rewards_arr = self.db.get_node(f'/{LOGS_GROUP}/{self.run_id}/{LOG_REWARDS_ARRAY}')
        rewards_arr.append(np.array([reward]))

    def _begin_new_trajectory(self):
        n_states = self.db.get_node(f'/{LOGS_GROUP}/{self.run_id}/{LOG_STATES_GROUP}/{STATE_ELEM}{0}').nrows
        n_transitions = self.db.get_node(f'/{LOGS_GROUP}/{self.run_id}/{LOG_REWARDS_ARRAY}').nrows
        self.db.get_node(f'/{LOGS_GROUP}/{self.run_id}/{TRAJECTORIES_ARRAY}').append((n_states, n_transitions))

    @staticmethod
    def _generate_run_id() -> str:
        return LOG_ID_GROUP + str(time.time()).replace('.', '_')

    def _initialize_monitor_run(self):
        root_children = [v._v_name for v in self.db.get_node('/')]
        if LOGS_GROUP not in root_children:
            self.db.create_group('/', LOGS_GROUP)
        log_root = self.db.create_group(f'/{LOGS_GROUP}', f"{self.run_id}")

        # create inner structure to store the data of the current run:
        self.db.create_vlarray(
            log_root,
            TRAJECTORIES_ARRAY,
            tb.UInt64Atom(shape=(2,), dflt=0),
            'An array of the indexes of the first state of each trajectory'
        )

        log_states_group = self.db.create_group(log_root, LOG_STATES_GROUP)
        for i, state_elem_shape in enumerate(self.state_shape):
            self.db.create_vlarray(
                log_states_group,
                f"{STATE_ELEM}{i}",
                tb.Atom.from_dtype(np.dtype((np.float32, state_elem_shape))),
                f'Element #{i} of the observation (state)',
                expectedrows=self.total_steps_estimate
            )

        action_atom = tb.UInt64Atom(shape=self.action_shape, dflt=0) if self.is_action_discrete else \
            tb.Atom.from_dtype(np.dtype((np.float32, self.action_shape)))
        self.db.create_vlarray(
            log_root,
            LOG_ACTIONS_ARRAY,
            atom=action_atom
        )

        self.db.create_vlarray(
            log_root,
            LOG_REWARDS_ARRAY,
            atom=tb.Float32Atom(shape=(), dflt=0.0)
        )


class DBReader:
    def __init__(self, db_file: str = 'monitor_db.h5'):
        self.db_file = db_file

    @property
    def db(self):
        return tb.open_file(self.db_file, mode='r')

    def get_logs_ids(self) -> List[str]:
        with self.db as db:
            logs_node = db.get_node(f'/{LOGS_GROUP}')
            return [v._v_name for v in logs_node._f_list_nodes()]

    def get_num_of_trajectories(self, log_id: str):
        if self._get_num_of_states(log_id=log_id) == 0:
            return 0

        with self.db as db:
            n_traj_rows = db.get_node(f'/{LOGS_GROUP}/{log_id}/{TRAJECTORIES_ARRAY}').nrows
        return n_traj_rows + 1

    def get_trajectories_lengths(self, log_id: str) -> List[int]:
        """
        Useful to know how many trajectories are there, and their lengths.
        Here, trajectory length is defined as the number of transitions (= number of actions).
        :return:
        """
        n_states = self._get_num_of_states(log_id=log_id)
        if n_states == 0:
            return []

        trajectories_start_indices = self._get_log_trajectories(log_id=log_id)
        trajectories_lengths = []
        for i in range(len(trajectories_start_indices)-1):
            trajectories_lengths.append((trajectories_start_indices[i+1][0]-1) - trajectories_start_indices[i][0])
        trajectories_lengths.append((n_states-1) - trajectories_start_indices[-1][0])

        return trajectories_lengths

    def get_transition_data(self, log_id: str, trajectory_index: int, transition_index: int):
        trajectories_start_indices = self._get_log_trajectories(log_id=log_id)

        with self.db as db:
            states_group_node = db.get_node(f'/{LOGS_GROUP}/{log_id}/{LOG_STATES_GROUP}')
            state_elements_nodes = states_group_node._f_list_nodes()
            state_idx = trajectories_start_indices[trajectory_index][0] + transition_index
            state = tuple(s[state_idx] for s in state_elements_nodes)
            next_state = tuple(s[state_idx+1] for s in state_elements_nodes)

            action_idx = trajectories_start_indices[trajectory_index][1] + transition_index
            action = db.get_node(f'/{LOGS_GROUP}/{log_id}/{LOG_ACTIONS_ARRAY}')[action_idx]
            reward = db.get_node(f'/{LOGS_GROUP}/{log_id}/{LOG_REWARDS_ARRAY}')[action_idx]

        return state, action, reward, next_state

    def get_trajectory_rewards(self, log_id: str, trajectory_index: int):
        trajectories_start_indices = self._get_log_trajectories(log_id=log_id)
        trajectories_lengths = self.get_trajectories_lengths(log_id=log_id)
        if len(trajectories_lengths) == 0:
            return None
        with self.db as db:
            start_index = trajectories_start_indices[trajectory_index][1]
            end_index = trajectories_start_indices[trajectory_index][1] + trajectories_lengths[trajectory_index]
            rewards = db.get_node(f'/{LOGS_GROUP}/{log_id}/{LOG_REWARDS_ARRAY}')[start_index:end_index]
            return rewards

    def get_num_of_state_elements(self, log_id: str):
        with self.db as db:
            return db.get_node(f'/{LOGS_GROUP}/{log_id}/{LOG_STATES_GROUP}')._v_nchildren

    def _get_num_of_states(self, log_id: str):
        with self.db as db:
            states_group_node = db.get_node(f'/{LOGS_GROUP}/{log_id}/{LOG_STATES_GROUP}')
            state_elements_nodes = states_group_node._f_list_nodes()
            if len(state_elements_nodes) == 0:
                return 0

            return db.get_node(f'/{LOGS_GROUP}/{log_id}/{LOG_STATES_GROUP}/{STATE_ELEM}{0}').nrows

    def _get_log_trajectories(self, log_id: str) -> List[Tuple[int, int]]:
        if self._get_num_of_states(log_id=log_id) == 0:
            return []

        trajectories = [(0, 0)]
        with self.db as db:
            stored_data = db.get_node(f'/{LOGS_GROUP}/{log_id}/{TRAJECTORIES_ARRAY}').read()
            if len(stored_data) > 0:
                trajectories.extend(np.array(stored_data, dtype=int).squeeze(1))
        return trajectories




import datetime

import gym
import tensorflow as tf
import os
from abc import ABC, abstractmethod
from typing import List
from copy import deepcopy

from replay_monitor.db import DBWriter, PosInTrajectory


class PerformanceMetric(ABC):
    @abstractmethod
    def track(self, *args, **kwargs):
        pass

    @property
    @abstractmethod
    def value(self):
        """
        Compute and return the value of the performance metric.
        :return:
        """
        pass

    @abstractmethod
    def log_metric_to_tf_summary(self):
        """
        Assume this method will be called with a default TF Summary Writer, output the relevant tf.summary calls.
        :return:
        """
        pass

    @abstractmethod
    def reset_state(self):
        pass

    def new_episode(self):
        """
        Optional. A method that is being called whenever a new episode starts
        :return:
        """
        pass


class PerStepPerformanceMetric(PerformanceMetric):
    """
    This class is intended for metrics that rely on 'per step' data such as observation, action, reward, info, etc.
    """

    @abstractmethod
    def track(self, observation, action, next_observation, reward, done, info, *args, **kwargs):
        pass


def get_space_shape(space: gym.spaces.Space):
    if isinstance(space, gym.spaces.Tuple):
        return tuple(get_space_shape(space) for space in space.spaces)
    else:
        return space.shape


DEFAULT_LOGS_DIR = 'monitor-logs'
DEFAULT_DB_FILENAME = 'monitor_db.h5'


class Monitor(gym.core.Wrapper):
    """
    This class is intended to provide a convenient way of tracking the training / running process of a Reinforcement
    Learning algorithm over an environment that follows the OpenAI's Gym conventions.
    """
    # TODO: Support Vector Envs.

    def __init__(self,
                 env: gym.Env,
                 performance_metrics: List[PerformanceMetric] = None,
                 log_to_db: bool = False,
                 logging_directory: str = DEFAULT_LOGS_DIR,
                 *arg, **kwargs):
        # initialize the directory for the logs:
        self.run_id = datetime.datetime.now().strftime("%d%m%Y-%H%M%S")
        logging_path = self._generate_logging_path(logging_directory=logging_directory, run_id=self.run_id)

        self.summary_writer = tf.summary.create_file_writer(logging_path)
        self.performance_metrics: List[PerformanceMetric] = performance_metrics \
            if performance_metrics is not None else []
        self.current_observation = None

        self.log_to_db = log_to_db
        self.db_file_path = os.path.join(logging_directory, DEFAULT_DB_FILENAME)
        self._db_writer = DBWriter(
            state_shape=get_space_shape(env.observation_space),
            action_shape=get_space_shape(env.action_space),
            is_action_discrete=isinstance(env.action_space, (gym.spaces.Discrete, gym.spaces.MultiDiscrete)),
            db_file=self.db_file_path,
            total_steps_estimate=None
        )
        self._last_observation = None
        self._is_first_observation = False

        super().__init__(env=env)

    def add_performance_metrics(self, performance_metrics: List[PerformanceMetric]):
        self.performance_metrics.extend(performance_metrics)

    def step(self, action):
        next_observation, reward, done, info = self.env.step(action)
        if self.log_to_db:
            # determine position in trajectory:
            self._db_writer.store_transition(
                state=self._last_observation,
                action=action,
                reward=reward,
                next_state=next_observation,
                info=info,
                pos_in_trajectory=self._determine_pos_in_trajectory(done)
            )
        self._is_first_observation = False

        self._track_performance(action=action, next_observation=next_observation, reward=reward, done=done, info=info)
        self.current_observation = deepcopy(next_observation)
        return next_observation, reward, done, info

    def reset(self, **kwargs):
        observation = self.env.reset(**kwargs)

        self._last_observation = observation
        self._is_first_observation = True

        self.current_observation = deepcopy(observation)
        for metric in self.performance_metrics:
            metric.new_episode()
        return observation

    def log_performance_metrics(self):
        with self.summary_writer.as_default():
            for metric in self.performance_metrics:
                metric.log_metric_to_tf_summary()

    def log_hparams(self, hparams: dict):
        """
        Logs the hyper parameters to a text table in Tensorboard.
        :param hparams: dictionary where the keys are strings with the names of the hyper parameters and the values
        are the corresponding values of the hyper parameters.
        """
        with self.summary_writer.as_default():
            tf.summary.text(name='Hyper Parameters', data=self.generate_hparams_md_table(hparams), step=0)

    def reset_metrics_states(self):
        for metric in self.performance_metrics:
            metric.reset_state()

    def _track_performance(self, action, next_observation, reward, done, info):
        for metric in self.performance_metrics:
            metric.track(self.current_observation,
                         action,
                         next_observation,
                         reward,
                         done,
                         info)

    @staticmethod
    def _generate_logging_path(logging_directory: str, run_id: str):
        logging_path = os.path.join(logging_directory, f'run_{run_id}')
        if not os.path.exists(logging_path):
            os.makedirs(logging_path)
        return logging_path

    @staticmethod
    def generate_hparams_md_table(hparams: dict):
        header = '| Hyper Parameter Name | Value | \n | --- | --- | \n'
        lines = [f'| {hp_k} | {hp_v} | \n' for hp_k, hp_v in hparams.items()]
        return header + ' '.join(lines)

    def _determine_pos_in_trajectory(self, done) -> PosInTrajectory:
        if self._is_first_observation:
            if done:
                pos = PosInTrajectory.ONLY
            else:
                pos = PosInTrajectory.FIRST
        else:
            if done:
                pos = PosInTrajectory.LAST
            else:
                pos = PosInTrajectory.MID
        return pos




import os
from enum import Enum, auto
from typing import Tuple

import numpy as np
from bokeh.models import Panel, Tabs, Slider, ColumnDataSource, Select, Spinner, Div
from bokeh.plotting import figure
from bokeh.layouts import column, row
from bokeh.server.server import Server

from replay_monitor.db import DBReader

from .monitor import DEFAULT_LOGS_DIR, DEFAULT_DB_FILENAME


def create_state_data_dict_from_state(state: Tuple[np.ndarray]):
    state_data_dicts = []
    for element in state:
        if element.shape[0] == 1:
            element = element[0]
        data = {}
        plot_type = determine_state_element_plot_type(element)
        if plot_type == StateElementPlotType.BAR:
            values = element.flatten()
            data['x'] = list(range(len(values)))
            data['state_element'] = values
            # data['state_x_range_1d'] = range(element.shape[0])
        elif plot_type == StateElementPlotType.MATRIX:
            data['state_element'] = [element]
            # data['state_x_range_2d'] = element.shape[1]
        elif plot_type == StateElementPlotType.COLOR_IMAGE:
            assert element.shape[2] == 3  # Assume color channel is last!
            xdim, ydim = element.shape[0:2]
            img = np.empty((xdim, ydim), dtype=np.uint32)
            view = img.view(dtype=np.uint8).reshape((xdim, ydim, 4))
            view[:, :, :-1] = np.flipud(element)
            view[:, :, -1] = 255
            data['state_element'] = [img]
        state_data_dicts.append(data)
    return state_data_dicts


def create_rewards_data_dict(rewards: np.ndarray):
    return {
        'rewards': rewards,
        'x': list(range(len(data_manager.trajectory_rewards)))
    }


def step_slider_change_handler(attr, old, new):
    global data_manager, ui_manager
    data_manager.change_transition(new)

    ui_manager.step_spinner.value = new
    ui_manager.step_slider.value = new
    ui_manager.update_ui_due_to_transition_change()


def trajectory_changed_handler(attr, old, new):
    global data_manager, ui_manager
    data_manager.change_transition(0, new)

    ui_manager.update_ui_due_to_trajectory_change()


def log_select_change_handler(attr, old, new):
    global data_manager, ui_manager
    data_manager.change_transition(0, 0, new)
    ui_manager.update_ui_due_to_log_change()


class StateElementPlotType(Enum):
    BAR = auto()
    MATRIX = auto()
    COLOR_IMAGE = auto()


def determine_state_element_plot_type(state_elem: np.ndarray) -> StateElementPlotType:
    if len(state_elem.shape) > 1 and state_elem.shape[0] == 1:
        state_elem = state_elem[0]

    if len(state_elem.shape) == 1:
        return StateElementPlotType.BAR
    elif len(state_elem.shape) == 2:
        if 1 in state_elem.shape:
            return StateElementPlotType.BAR
        else:
            return StateElementPlotType.MATRIX
    elif len(state_elem.shape) == 3 and state_elem.shape[2] == 3:
        return StateElementPlotType.COLOR_IMAGE

def create_state_layout(state, data_sources, title):
    tabs_list = []
    for i, element in enumerate(state):
        if len(element.shape) > 1 and element.shape[0] == 1:
            element = element[0]

        plot_type = determine_state_element_plot_type(element)
        if plot_type == StateElementPlotType.BAR:
            fig = figure(plot_width=600, plot_height=600)
            fig.vbar(x='x', width=0.5, bottom=0, top='state_element', source=data_sources[i])
        elif plot_type == StateElementPlotType.MATRIX:
            range_max = max(element.shape[1], element.shape[0])
            fig = figure(plot_width=600, plot_height=600, x_range=(0, range_max), y_range=(0, range_max))
            fig.image(image='state_element', x=0, y=0, dw=element.shape[1], dh=element.shape[0], palette="Spectral11",
                      source=data_sources[i])
        elif plot_type == StateElementPlotType.COLOR_IMAGE:
            range_max = max(element.shape[1], element.shape[0])
            fig = figure(plot_width=600, plot_height=600, x_range=(0, range_max), y_range=(0, range_max))
            fig.image_rgba(image='state_element', x=0, y=0, dw=element.shape[1], dh=element.shape[0],
                           source=data_sources[i])
        else:
            continue

        tabs_list.append(Panel(child=fig, title=f"State Element {i}"))
    state_layout = column(
        Div(text=f"<h2>{title}</h2>"),
        Tabs(tabs=tabs_list)
    )
    return state_layout


class DataManager:
    def __init__(self, db_file_path: str):
        self.db_file_path = db_file_path
        db_reader = DBReader(db_file=db_file_path)
        self.log_ids = db_reader.get_logs_ids()
        self.current_log = self.log_ids[0]
        self.trajectory_index = 0
        self.transition_index = 0
        self.trajectories_lengths = db_reader.get_trajectories_lengths(self.current_log)
        self.n_state_elements = db_reader.get_num_of_state_elements(self.current_log)

        self.trajectory_rewards = db_reader.get_trajectory_rewards(log_id=self.current_log,
                                                                   trajectory_index=self.trajectory_index)

        self.s, self.a, self.r, self.s2 = self.get_transition(self.current_log, self.trajectory_index,
                                                              self.transition_index)

        self.state_elements_sizes = [state_elem.size for state_elem in self.s]

    def get_transition(self, log_id: str, trajectory_index: int, transition_index: int):
        db_reader = DBReader(db_file=self.db_file_path)
        data = db_reader.get_transition_data(log_id, trajectory_index=trajectory_index,
                                             transition_index=transition_index)
        return data

    def change_transition(self, transition_index: int, trajectory_index: int = None, log_id: str = None):
        is_log_changed = False
        if log_id is not None:
            is_log_changed = log_id != self.current_log
            self.current_log = log_id

        if trajectory_index is not None:
            is_different_trajectory = self.trajectory_index != trajectory_index
            self.trajectory_index = trajectory_index
            if is_different_trajectory or is_log_changed:
                db_reader = DBReader(db_file=self.db_file_path)
                self.trajectory_rewards = db_reader.get_trajectory_rewards(log_id=self.current_log,
                                                                           trajectory_index=self.trajectory_index)

        self.transition_index = transition_index

        self.s, self.a, self.r, self.s2 = self.get_transition(self.current_log, self.trajectory_index,
                                                              self.transition_index)

        if log_id is not None and is_log_changed:
            db_reader = DBReader(db_file=self.db_file_path)
            self.trajectories_lengths = db_reader.get_trajectories_lengths(self.current_log)
            self.n_state_elements = db_reader.get_num_of_state_elements(self.current_log)
            self.state_elements_sizes = [state_elem.size for state_elem in self.s]

    def get_current_trajectory_length(self):
        return self.trajectories_lengths[self.trajectory_index]


class UIManager:
    def __init__(self, data_manager: DataManager):
        self.data_manager = data_manager

        self.log_select = Select(title="Choose a log:", value=data_manager.current_log, options=data_manager.log_ids)
        self.log_select.on_change('value', log_select_change_handler)

        self.step_slider = None
        self._generate_step_slider()

        self.trajectory_select = None
        self._generate_trajectory_select()

        self.state_data_sources = None
        self.next_state_data_sources = None
        self.state_layout = None
        self.next_state_layout = None
        self.states_layout = None
        self._create_states_ui()

        self.rewards_data_source = ColumnDataSource(data=create_rewards_data_dict(self.data_manager.trajectory_rewards))
        self.trajectory_rewards_fig = None
        self._create_trajectory_rewards_ui()

        self.textual_action_reward = Div(text=self._generate_textual_action_reward_str())

        self.states_layout = row(self.state_layout, self.next_state_layout)
        self.sliders_layout = column(row(self.step_slider, self.step_spinner), self.trajectory_select, sizing_mode='stretch_width')
        self.layout = column(self.log_select,
                             self.trajectory_rewards_fig,
                             self.textual_action_reward,
                             self.states_layout,
                             self.sliders_layout)

    def update_ui_due_to_log_change(self):
        self._create_states_ui()
        self.states_layout.children = [self.state_layout, self.next_state_layout]

        self.rewards_data_source.stream(new_data=create_rewards_data_dict(self.data_manager.trajectory_rewards),
                                        rollover=len(self.data_manager.trajectory_rewards))

        self.trajectory_select.high = len(self.data_manager.trajectories_lengths)-1
        self.trajectory_select.value = self.data_manager.trajectory_index

        self.step_slider.value = self.data_manager.transition_index
        self.step_slider.end = self.data_manager.get_current_trajectory_length()-1
        self.step_spinner.high = self.step_slider.end
        self.step_spinner.value = self.step_slider.value

    def update_ui_due_to_transition_change(self):
        state_data_dicts = create_state_data_dict_from_state(self.data_manager.s)
        next_state_data_dicts = create_state_data_dict_from_state(self.data_manager.s2)
        for i, element in enumerate(self.data_manager.s):
            self.state_data_sources[i].stream(new_data=state_data_dicts[i],
                                                    rollover=self.data_manager.state_elements_sizes[i])
            self.next_state_data_sources[i].stream(new_data=next_state_data_dicts[i],
                                                         rollover=self.data_manager.state_elements_sizes[i])

        self.textual_action_reward.text = self._generate_textual_action_reward_str()

    def update_ui_due_to_trajectory_change(self):
        self.update_ui_due_to_transition_change()

        self.rewards_data_source.stream(new_data=create_rewards_data_dict(self.data_manager.trajectory_rewards),
                                        rollover=len(self.data_manager.trajectory_rewards))

        self.step_slider.value = self.data_manager.transition_index
        self.step_slider.end = self.data_manager.get_current_trajectory_length() - 1
        self.step_spinner.high = self.step_slider.end
        self.step_spinner.value = self.step_slider.value

    def _generate_step_slider(self):
        self.step_slider = Slider(start=0, end=self.data_manager.get_current_trajectory_length()-1,
                                  value=self.data_manager.transition_index, step=1, title="Time Step",
                                  sizing_mode='stretch_width')
        self.step_spinner = Spinner(low=0, high=self.step_slider.end, step=1, value=self.step_slider.value, width=100)
        self.step_slider.on_change('value', step_slider_change_handler)
        self.step_spinner.on_change('value', step_slider_change_handler)

    def _generate_trajectory_select(self):
        self.trajectory_select = Spinner(
            title="Choose a trajectory:",
            low=0,
            high=len(self.data_manager.trajectories_lengths)-1,
            step=1,
            value=self.data_manager.trajectory_index,
        )
        self.trajectory_select.on_change('value', trajectory_changed_handler)

    def _create_states_ui(self):
        self.state_data_sources = [ColumnDataSource(data=data_dict)
                                   for data_dict in create_state_data_dict_from_state(self.data_manager.s)]
        self.next_state_data_sources = [ColumnDataSource(data=data_dict)
                                        for data_dict in create_state_data_dict_from_state(self.data_manager.s2)]
        self.state_layout = create_state_layout(self.data_manager.s, self.state_data_sources, 'State')
        self.next_state_layout = create_state_layout(self.data_manager.s2, self.next_state_data_sources, 'Next State')

    def _create_trajectory_rewards_ui(self):
        fig = figure(plot_width=600, plot_height=300, sizing_mode='stretch_width')
        fig.circle(x='x', y='rewards', source=self.rewards_data_source, name='trajectory_rewards_plot')
        self.trajectory_rewards_fig = fig
        return fig

    def _generate_textual_action_reward_str(self):
        text = f'<h4>Reward: {self.data_manager.r[0]}<br>Action: {self.data_manager.a[0]}</h4>'
        return text


db_file_path = os.path.join(DEFAULT_LOGS_DIR, DEFAULT_DB_FILENAME)
data_manager = None
ui_manager = None


def start_app(doc):
    global data_manager, ui_manager, db_file_path
    data_manager = DataManager(db_file_path)
    ui_manager = UIManager(data_manager=data_manager)
    doc.add_root(ui_manager.layout)


def _start_server(db_path: str = None):
    global db_file_path

    if db_path is not None:
        db_file_path = db_path

    server = Server({'/': start_app}, num_procs=1)
    server.start()

    print('Opening Bokeh application on http://localhost:5006/')

    server.io_loop.add_callback(server.show, "/")
    server.io_loop.start()


def start_server():
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument('--db_path', required=False)
    args = parser.parse_args()

    _start_server(args.db_path)


if __name__ == '__main__':
    start_server()


import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="replay-monitor", # Replace with your own username
    version="0.0.5",
    author="Leor Cohen",
    author_email="liorcohen5@gmail.com",
    description="A tool for easy data exploration in reinforcement learning environments.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/liorcohen5/replay-monitor",
    download_url="https://github.com/liorcohen5/replay-monitor/archive/0.0.5.tar.gz",
    keywords=['reinforcement learning', 'tool', 'data exploration', 'replay', 'monitor', 'analytical tool'],
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=['gym', 'bokeh', 'tensorflow', 'numpy', 'tables'],
    entry_points={
        'console_scripts': ['replay-monitor=replay_monitor.server:start_server'],
    },
    python_requires='>=3.6',
)